#----------------------------------------------------------------------------------------------------
#Author: Warren Jacobus
#        3065044
#        Univeristy of the Western Cape
#----------------------------------------------------------------------------------------------------
from heapq import *
import heapq
from collections import defaultdict
from operator import itemgetter
from prioritydictionary import priorityDictionary
from graph import DiGraph
#----------------------------------------------------------------------------------------------------
#Description (algorithms.py):
#           This script has different algorithm implementations used in networking.
#           These algorithms get used to simulate different network scenarios.
#Inputs:
#           None.
#Outputs:
#           None.
#Caveats:
#           Described below
#----------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------
#prim:
#            A greedy algorithm that finds a minimum spanning tree for a
#            connected weighted undirected graph.
# 
#            -Initialize a tree with a single vertex, chosen arbitrarily from the graph.
#            -Grow the tree by one edge: of the edges that connect the tree to vertices not yet in the tree,
#             find the minimum-weight edge, and transfer it to the tree.
#            -Repeat previous step until all vertices are in the tree.
#Inputs:
#            Takes a set of nodes and a set of corresponding edges, where edges is a tuple (nodeN,nodeM,cost)            
#Outputs:
#            Minimum spanning tree of graph
#Caveats:
#            nodes is required to be a list.    [node1,......,nodeK]
#            edges is required to be a 3-tuple. (nodeN,nodeM,cost)
#----------------------------------------------------------------------------------------------------
def prim( nodes, edges ): # Prim's Algorithm
    conn = defaultdict( list )
    for n1,n2,c in edges:
        conn[ n1 ].append( (c, n1, n2) )
        conn[ n2 ].append( (c, n2, n1) )
 
    mst = []
    used = set( nodes[ 0 ] )
    usable_edges = conn[ nodes[0] ][:]
    heapify( usable_edges )
 
    while usable_edges:
        cost, n1, n2 = heappop( usable_edges )
        if n2 not in used:
            used.add( n2 )
            mst.append( ( n1, n2, cost ) )
 
            for e in conn[ n2 ]:
                if e[ 2 ] not in used:
                    heappush( usable_edges, e )
    return mst
#----------------------------------------------------------------------------------------------------
#DisjointSet:
#            Partitions a set of items. 
#            A partition is a set of sets such that each item is in one and only one set.
#Inputs:
#            Has a required dictionary, by default           
#Outputs:
#            None.
#            Contains methods: add(item)
#                              find(item)
#                              union(item1,item2)
#Caveats:
#            There are no known limitations to date
#----------------------------------------------------------------------------------------------------
class DisjointSet(dict): # Used in Kruskal's Algorithm below
    def add(self, item):
        self[item] = item
 
    def find(self, item):
        parent = self[item]
 
        while self[parent] != parent:
            parent = self[parent]
 
        self[item] = parent
        return parent
 
    def union(self, item1, item2):
        self[item2] = self[item1]
#----------------------------------------------------------------------------------------------------
#kruskal:
#            A greedy algorithm that finds a minimum spanning tree for a
#            connected weighted graph. 
#
#            -Create a forest F (a set of trees), where each vertex in the graph is a separate tree.
#            -Create a set S containing all the edges in the graph.
#            -While S is nonempty and F is not yet spanning.
#               Remove an edge with minimum weight from S.
#               If that edge connects two different trees, then add it to the forest, combining two trees into a single tree.
#Inputs:
#            A set of nodes and corresponding edges.                      
#Outputs:
#            Minimum spanning tree of graph
#Caveats:
#            nodes is required to be a list.    [node1,......,nodeK]
#            edges is required to be a 3-tuple. (nodeN,nodeM,cost)
#----------------------------------------------------------------------------------------------------
def kruskal( nodes, edges ): # Kruskal's Algorithm
    forest = DisjointSet()
    mst = []
    for n in nodes:
        forest.add( n )
 
    size = len(nodes) - 1
 
    for e in sorted( edges, key=itemgetter( 2 )):
        n1, n2, _ = e
        t1 = forest.find(n1)
        t2 = forest.find(n2)
        if t1 != t2:
            mst.append(e)
            size -= 1
            if size == 0:
                return mst
         
            forest.union(t1, t2)
    return mst
#----------------------------------------------------------------------------------------------------
#  Modification of Kevin R <KRPent@gmail.com> 
## Computes K paths from a source to a sink in the supplied graph.
#
# @param graph A digraph of class Graph.
# @param start The source node of the graph.
# @param sink The sink node of the graph.
# @param K The amount of paths being computed.
#
# @retval [] Array of paths, where [0] is the shortest, [1] is the next 
# shortest, and so on.
#
def ksp_yen(graph, node_start, node_end, max_k=2):
    distances, previous = dijkstra(graph, node_start)
    
    A = [{'cost': distances[node_end], 
          'path': path(previous, node_start, node_end)}]
    B = []
    
    if not A[0]['path']: return A
    
    for k in range(1, max_k):
        for i in range(0, len(A[-1]['path']) - 1):
            node_spur = A[-1]['path'][i]
            path_root = A[-1]['path'][:i+1]
            
            edges_removed = []
            for path_k in A:
                curr_path = path_k['path']
                if len(curr_path) > i and path_root == curr_path[:i+1]:
                    cost = graph.remove_edge(curr_path[i], curr_path[i+1])
                    if cost == -1:
                        continue
                    edges_removed.append([curr_path[i], curr_path[i+1], cost])
            
            path_spur = dijkstra(graph, node_spur, node_end)
            
            if path_spur['path']:
                path_total = path_root[:-1] + path_spur['path']
                dist_total = distances[node_spur] + path_spur['cost']
                potential_k = {'cost': dist_total, 'path': path_total}
            
                if not (potential_k in B):
                    B.append(potential_k)
            
            for edge in edges_removed:
                graph.add_edge(edge[0], edge[1], edge[2])
        
        if len(B):
            B = sorted(B, key=itemgetter('cost'))
            A.append(B[0])
            B.pop(0)
        else:
            break
    
    return A
#----------------------------------------------------------------------------------------------------
#  Modification of Kevin R <KRPent@gmail.com>
## Computes the shortest path from a source to a sink in the supplied graph.
#
# @param graph A digraph of class Graph.
# @param node_start The source node of the graph.
# @param node_end The sink node of the graph.
#
# @retval {} Dictionary of path and cost or if the node_end is not specified,
# the distances and previous lists are returned.
#
def dijkstra(graph, node_start, node_end=None):
    distances = {}      
    previous = {}       
    Q = priorityDictionary()
    
    for v in graph:
        distances[v] = graph.INFINITY
        previous[v] = graph.UNDEFINDED
        Q[v] = graph.INFINITY
    
    distances[node_start] = 0
    Q[node_start] = 0
    
    for v in Q:
        if v == node_end: break

        for u in graph[v]:
            cost_vu = distances[v] + graph[v][u]
            
            if cost_vu < distances[u]:
                distances[u] = cost_vu
                Q[u] = cost_vu
                previous[u] = v

    if node_end:
        return {'cost': distances[node_end], 
                'path': path(previous, node_start, node_end)}
    else:
        return (distances, previous)
#----------------------------------------------------------------------------------------------------
#  Modification of Kevin R <KRPent@gmail.com>
## Finds a paths from a source to a sink using a supplied previous node list.
#
# @param previous A list of node predecessors.
# @param node_start The source node of the graph.
# @param node_end The sink node of the graph.
#
# @retval [] Array of nodes if a path is found, an empty list if no path is 
# found from the source to sink.
#
def path(previous, node_start, node_end):
    route = []

    node_curr = node_end    
    while True:
        route.append(node_curr)
        if previous[node_curr] == node_start:
            route.append(node_start)
            break
        elif previous[node_curr] == DiGraph.UNDEFINDED:
            return []
        
        node_curr = previous[node_curr]
    
    route.reverse()
    return route
#----------------------------------------------------------------------------------------------------
#END
#----------------------------------------------------------------------------------------------------