#----------------------------------------------------------------------------------------------------
#Author: Warren Jacobus
#        3065044
#        Univeristy of the Western Cape
#----------------------------------------------------------------------------------------------------
from math import sin, cos, radians, sqrt, asin,log
specs = open("device_specs/specs.txt",'r').read().strip("\n").split()
Ptx = float(specs[0])
Gtx = float(specs[1]) 
Ltx = float(specs[2])
Grx = float(specs[3])
Lrx = float(specs[4])
Rx  = float(specs[5])
frq = float(specs[6])
#----------------------------------------------------------------------------------------------------
#Description (compute.py):
#           Contains methods used to compute:
#               distances between each node,
#               lfs(free space loss between 2 nodes),
#               Prx(received power/link budget equation),
#               link_margin between 2 nodes,
#               FeasibleLinks between all nodes in graph,
#               Distances between all nodes in a graph,
#               FreeSpaceLoss(lfs) between all nodes in a graph, and
#               LinkMargins between all nodes in a graph.
#Inputs:
#           None.
#Outputs:
#           None.
#Caveats:
#           Described below
#----------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------
#distance:
#            Implementation of Haversine's algorithm.
#            Determines the distance between 2 gps coordinates using their latitudes and longtitudes.
#Inputs:
#            2 gps coordinates
#Outputs:
#            distance between 2 gps coordinates [km]
#Caveats:
#            Only makes use of the latitude and longtitude and not the altitude
#----------------------------------------------------------------------------------------------------
def distance(a, b):
   r = 6367442.5    # average earth radius in m

   dLat = radians(a[0]-b[0])
   dLon = radians(a[1]-b[1])

   x = sin(dLat/2) ** 2 + cos(radians(a[0])) * cos(radians(b[0])) * sin(dLon/2) ** 2
   y = 2 * asin(sqrt(x))

   d = r * y
   return d # in metres
#----------------------------------------------------------------------------------------------------
#LFS:
#            Determines the free space loss between 2 nodes
#Inputs:
#            Distance between 2 nodes
#Outputs:
#            free space loss between the 2 nodes
#Caveats:
#            There are no limitations to date.
#----------------------------------------------------------------------------------------------------
def LFS(d): # link budget calculation
  if d == 0.0:
    return 0.0
  return (32.45 + 20*log(frq,10) + 20*log(d,10))
#----------------------------------------------------------------------------------------------------
#Prx:
#            Determines the received power
#Inputs:
#            free space loss between 2 nodes
#Outputs:
#            received power
#Caveats:
#            There are no limitations to date
#----------------------------------------------------------------------------------------------------
def Prx(lfs): # Received Power (link budget equation)
  if lfs == 0.0:
    return 0.0
  return (Ptx + Gtx - Ltx - lfs + Grx - Lrx)
#----------------------------------------------------------------------------------------------------
#link_margin:
#            Determines the link margin between 2 nodes
#Inputs:
#            received power (Prx)
#Outputs:
#            link margin
#Caveats:
#            There are no limitations to date
#----------------------------------------------------------------------------------------------------
def link_margin(prx):
  if prx == 0.0:
    return 0.0
  return prx - Rx
#----------------------------------------------------------------------------------------------------
#FeasibleLinks:
#            Determines the feasible links for all locations
#Inputs:
#            link margins
#Outputs:
#            list of 3-tuples of feasible links
#Caveats:
#            There are no limitations to date.
#----------------------------------------------------------------------------------------------------
def FeasibleLinks(link_margins): # compute feasible links for all locations
   links = []
   feasible = []
   for i in xrange(1,len(link_margins)):
      for j in xrange(1,len(link_margins)):
         #if link_margins[i][j] >= 10.0 and j <= i:
         if link_margins[i][j] >= 10.0:
            links.append((link_margins[i][0],link_margins[0][j],round(link_margins[i][j],2)))
   return links
#----------------------------------------------------------------------------------------------------
#Distances:
#            Determines the distance between each node in a graph
#Inputs:
#            list places containing gps coordinates of each node and list placeName holding
#            the name of each node in the graph
#Outputs:
#            2d-array/list of all distances between every node in  a graph
#Caveats:
#            There are no limitations to date
#----------------------------------------------------------------------------------------------------
def Distances(places,placeName):
   # Prepare table
   distances = [[0]*(len(places)+1) for i in xrange(len(places)+1)]
   distances[0][0] = "PLACES"
   for i in xrange(1,len(distances)):
      distances[0][i] = placeName[i-1]
   for i in xrange(1,len(distances)):
      distances[i][0] = placeName[i-1]

   # Fill table
   for i in xrange(1,len(distances)):
      for j in xrange(1,len(distances)):
         distances[i][j] = (distance(places[i-1],places[j-1]))/1000.0 # convert m to km
   return distances
#----------------------------------------------------------------------------------------------------
#FreeSpaceLoss:
#            Determines the free space loss between each node in a graph
#Inputs:
#            list distances,places, and placeName
#Outputs:
#            2d-array/list of free space loss between every node in a graph
#Caveats:
#            Only makes use of the latitude and longtitude and not the altitude
#----------------------------------------------------------------------------------------------------
def FreeSpaceLoss(distances,places,placeName):
   # Prepare table
   lfs = [[0]*(len(places)+1) for i in xrange(len(places)+1)]
   lfs[0][0] = "PLACES"
   for i in xrange(1,len(lfs)):
      lfs[0][i] = placeName[i-1]
   for i in xrange(1,len(lfs)):
      lfs[i][0] = placeName[i-1]

   # Fill table
   for i in xrange(1,len(lfs)):
      for j in xrange(1,len(lfs[0])):
         lfs[i][j] = LFS(distances[i][j])
   return lfs
#----------------------------------------------------------------------------------------------------
#LinkMargins:
#            Determines the link margins between every node in a graph
#Inputs:
#            list lfs(free space loss),places,placeName
#Outputs:
#            2d-array/list of link margins
#Caveats:
#            There are no limitations to date
#----------------------------------------------------------------------------------------------------
def LinkMargins(lfs,places,placeName): # compute link margin between each node
   # Prepare table
   link_margins = [[0]*(len(places)+1) for i in xrange(len(places)+1)]
   link_margins[0][0] = "PLACES"
   for i in xrange(1,len(lfs)):
      link_margins[0][i] = placeName[i-1]
   for i in xrange(1,len(lfs)):
      link_margins[i][0] = placeName[i-1]

   # Fill table
   for i in xrange(1,len(link_margins)):
      for j in xrange(1,len(link_margins[0])):
         link_margins[i][j] = link_margin(Prx(lfs[i][j]))
   return link_margins
#----------------------------------------------------------------------------------------------------
#END
#----------------------------------------------------------------------------------------------------
