#----------------------------------------------------------------------------------------------------
#Author: Warren Jacobus
#        3065044
#        Univeristy of the Western Cape
#----------------------------------------------------------------------------------------------------
import networkx as nx
import matplotlib.pyplot as plt
import ast
import algorithms
#----------------------------------------------------------------------------------------------------
#Description (draw_graph.py):
#           Visualizes the graphs of the mesh generated along with the mst's
#           generated from Prim and Kruskal of the mesh.
#Inputs:
#           None.
#Outputs:
#           pdf images of a mesh and the kruskal and prim's graphs from it.
#Caveats:
#           Described below.
#----------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------
#draw_graph:
#            Visualizes the graphs of the mst's of prim and kruskal generated
#            from the mesh network and visualizes the mesh too.
#Inputs:
#            graph(the mesh network), optional labels for edges, the graph layout,
#            node sizes, node color, node alpha(text opaqueness), node_text_size,
#            edge_color, edge_alpha(text opaqueness), edge_tickness, edge_text_position,
#            and text_font
#Outputs:
#            3 graphs and pdf images thereof. (mesh.pdf,kruskal_graph.pdf,prim_graph.pdf)
#Caveats:
#            Only shell-layout gives usable results, other layouts generate messy graphs.
#----------------------------------------------------------------------------------------------------
def draw_graph(graph, labels=None, graph_layout='shell',
               node_size=1600, node_color='red', node_alpha=0.4,
               node_text_size=10,
               edge_color='blue', edge_alpha=0.4, edge_tickness=1,
               edge_text_pos=0.3,
               text_font='sans-serif'):

    # Hide axis on graph
    plt.gca().axes.get_xaxis().set_visible(False)
    plt.gca().axes.get_yaxis().set_visible(False)

    # read mesh and locations from files to enables the computations of prim and kruskal (automation purposes)
    f = open('gps_cords/locations.txt','r')
    g = open('graphs/text/graph.txt','r')

    locations = [(i.strip("\n")).split() for i in f.readlines()]
    nodes = [i[0] for i in locations]
    edges = [ast.literal_eval(i) for i in g.read().split('\n')[:-1][:]]

    # hack to fool heap(built_in min heap) into being a max heap
    # did not want to change comparison operator in source code because I might forget to change it back 
    edges_for_prim_kruskal = [(item[0],item[1],item[2]**-1) for item in edges]

    f.close()
    g.close()

    # Compute prim and kruskal graphs
    primm = algorithms.prim( nodes, edges_for_prim_kruskal)
    krusk = algorithms.kruskal(nodes, edges_for_prim_kruskal)

    # convert cost back to original value
    primm = [(item[0],item[1],round(item[2]**-1,2)) for item in primm] 
    krusk = [(item[0],item[1],round(item[2]**-1),2) for item in krusk]

    # Put prim and kruskal graph in proper format for visualization
    primm_graph = [(i[0],i[1]) for i in primm]
    krusk_graph = [(i[0],i[1]) for i in krusk]

    # create networkx graph
    G=nx.Graph()
    G2=nx.Graph()
    G3=nx.Graph() 

    # add edges
    for edge in graph:
        G.add_edge(edge[0], edge[1])
        G2.add_edge(edge[0],edge[1],color="blue")
        G3.add_edge(edge[0],edge[1],color="green")

    # these are different layouts for the network you may try
    # shell seems to work best
    if graph_layout == 'spring':
        graph_pos=nx.spring_layout(G)
        graph_pos2=nx.spring_layout(G2)
        graph_pos3=nx.spring_layout(G3)
    elif graph_layout == 'spectral':
        graph_pos=nx.spectral_layout(G)
        graph_pos2=nx.spectral_layout(G2)
        graph_pos3=nx.spectral_layout(G3)
    elif graph_layout == 'random':
        graph_pos=nx.random_layout(G)
        graph_pos2=nx.random_layout(G2)
        graph_pos3=nx.random_layout(G3)
    else:
        graph_pos=nx.shell_layout(G)
        graph_pos2=nx.shell_layout(G2)
        graph_pos3=nx.shell_layout(G3)

    # draw graph
    plt.figure("Original Mesh")
    nx.draw_networkx_nodes(G,graph_pos,node_size=node_size, 
                           alpha=node_alpha, node_color=node_color)
    nx.draw_networkx_edges(G,graph_pos,width=edge_tickness,
                           alpha=edge_alpha,edge_color=edge_color)
    nx.draw_networkx_labels(G, graph_pos,font_size=node_text_size,
                            font_family=text_font)
    

    if labels is None:
        labels = range(len(graph))

    edge_labels = dict(zip(graph, labels))
    edge_labels2 = dict(zip(primm_graph,primm))
    edge_labels3 = dict(zip(krusk_graph,krusk))


    #nx.draw_networkx_edge_labels(G, graph_pos, edge_labels=edge_labels, 
    #                             label_pos=edge_text_pos, font_size = 1)
    # Draw mesh    
    plt.axis('off')
    plt.savefig("graphs/images/png/mesh.png",facecolor="white")
    #plt.show()

    # Draw Prim
    plt.figure("Prim")
    nx.draw_networkx_nodes(G2,graph_pos2,node_size=node_size, 
                           alpha=node_alpha, node_color=node_color)

    # Highlight prim and kruskal
    nx.draw_networkx_edges(G2,graph_pos2,edgelist=primm,width=edge_tickness,alpha=edge_alpha,
                            edge_color='g')
    nx.draw_networkx_labels(G2, graph_pos2,font_size=node_text_size,
                            font_family=text_font)
    #nx.draw_networkx_edge_labels(G2, graph_pos2, edge_labels=edge_labels2, 
    #                             label_pos=edge_text_pos, font_size = 4)

    # show graph
    plt.axis('off')
    plt.savefig("graphs/images/png/prim_graph.png",facecolor="white")
    #plt.show()

    # Draw Kruskal
    plt.figure("Kruskal")
    nx.draw_networkx_nodes(G3,graph_pos3,node_size=node_size, 
                           alpha=node_alpha, node_color=node_color)
    nx.draw_networkx_edges(G3,graph_pos3,edgelist=krusk,width=edge_tickness,alpha=edge_alpha,
                            edge_color='b')
    nx.draw_networkx_labels(G3, graph_pos3,font_size=node_text_size,
                            font_family=text_font)
    #nx.draw_networkx_edge_labels(G3, graph_pos3, edge_labels=edge_labels3, 
    #                             label_pos=edge_text_pos, font_size = 4)

    plt.axis('off')
    plt.savefig("graphs/images/png/kruskal_graph.png",facecolor="white")
    #plt.show()
#----------------------------------------------------------------------------------------------------
#END
#----------------------------------------------------------------------------------------------------
